tool
extends EditorPlugin

const print_prefix = "Save as STL | "
var eds = get_editor_interface().get_selection()
var save_dialog
var save_path = ""
var selected_node = null

func _enter_tree():
	add_tool_menu_item("Save Selected Node's Mesh As STL", self, "do_export")

func add_dialog():
	save_dialog = FileDialog.new()
	save_dialog.access = FileDialog.ACCESS_FILESYSTEM
	save_dialog.filters = ["*.stl"]
	save_dialog.resizable = true
	save_dialog.margin_right = 500
	save_dialog.margin_bottom = 300
	save_dialog.connect("file_selected", self, "_on_file_selected")
	add_child(save_dialog)

func do_export(ud):
	var selection = eds.get_selected_nodes()
	if selection.empty():
		push_error(print_prefix+"Error: No node selected")
		return
	
	selected_node = selection[0]
	if not selected_node is MeshInstance and not selected_node is CSGShape:
		push_error(print_prefix+"Error: Selected node is not a MeshInstance or a CSGShape")
		selected_node = null
		return
	
	if save_dialog == null:
		add_dialog()
	if save_path:
		save_dialog.current_path = save_path
	save_dialog.current_file = "%s.stl" % selected_node.name
	save_dialog.window_title = "Save %s's Mesh as STL" % selected_node.name
	save_dialog.popup_centered(Vector2())


func save_to_stl():
	var am
	if selected_node is CSGShape:
		print(print_prefix+"Info: CSGShape detected")
		selected_node._update_shape()
		var m = selected_node.get_meshes()#[1]
		if m.empty():
			push_error(print_prefix+"Error: Empty Shape")
			return
		am = m[1]
	elif selected_node is MeshInstance:
		print(print_prefix+"Info: MeshInstance detected")
		var m = selected_node.mesh
		if m is PrimitiveMesh:
			print(print_prefix+"Info: PrimitiveMesh detected")
			am = ArrayMesh.new()
			am.add_surface_from_arrays(Mesh.PRIMITIVE_TRIANGLES, m.get_mesh_arrays())
		elif m is ArrayMesh:
			print(print_prefix+"Info: ArrayMesh detected")
			am = m
		else:
			push_error(print_prefix+"Error: Unknown Mesh Type")
			return
	else:
		push_error(print_prefix+"Error: Unknown Node Type")
		return
	
	var sname = save_path.get_file().get_basename()
	var sfile = File.new()
	sfile.open(save_path+".part", File.WRITE)
	sfile.store_string("solid %s\n" % sname)
	
	for sufid in am.get_surface_count():
		if not am.surface_get_primitive_type(sufid) == Mesh.PRIMITIVE_TRIANGLES:
			push_warning(print_prefix+"Warning: surface id "+str(sufid)+" ignored (only PRIMITIVE_TRIANGLES supported)")
			continue
		
		var sufarrs = am.surface_get_arrays(sufid)
		
		if am.surface_get_array_index_len(sufid):
			print(print_prefix+"Info: Indexed Mode detected")
			for n in range(0, am.surface_get_array_index_len(sufid), 3):
				var i0 = sufarrs[ArrayMesh.ARRAY_INDEX][n+0]
				var i1 = sufarrs[ArrayMesh.ARRAY_INDEX][n+1]
				var i2 = sufarrs[ArrayMesh.ARRAY_INDEX][n+2]
				var a = turn(sufarrs[ArrayMesh.ARRAY_VERTEX][i0])
				var b = turn(sufarrs[ArrayMesh.ARRAY_VERTEX][i1])
				var c = turn(sufarrs[ArrayMesh.ARRAY_VERTEX][i2])
				var norm = get_triangle_normal(a, c, b)
				write_facet(sfile, norm, a, c, b)
		else:
			print(print_prefix+"Info: Normal Mode assumed")
			for i in range(0, am.surface_get_array_len(sufid), 3):
				var a = turn(sufarrs[ArrayMesh.ARRAY_VERTEX][i+0])
				var b = turn(sufarrs[ArrayMesh.ARRAY_VERTEX][i+1])
				var c = turn(sufarrs[ArrayMesh.ARRAY_VERTEX][i+2])
				var norm = get_triangle_normal(a, c, b)
				write_facet(sfile, norm, a, c, b)
	
	sfile.store_string("endsolid %s\n" % sname)
	sfile.close()
	Directory.new().rename(save_path+".part", save_path)


func turn(a):
	var m = Basis()
	m = m.rotated(Vector3(1,0,0), PI/2)
	return m * a

func get_triangle_normal(a, b, c):
	var side1 = b - a
	var side2 = c - a
	var normal = side1.cross(side2)
	return normal

func write_facet(sfile, norm, a, b, c):
	sfile.store_string("facet normal %f %f %f\nouter loop\n" % [norm.x, norm.y, norm.z])
	sfile.store_string("vertex %f %f %f\n" % [a.x, a.y, a.z])
	sfile.store_string("vertex %f %f %f\n" % [b.x, b.y, b.z])
	sfile.store_string("vertex %f %f %f\n" % [c.x, c.y, c.z])
	sfile.store_string("endloop\nendfacet\n")


func _on_file_selected(path):
	save_path = path
	save_to_stl()


func _exit_tree():
	remove_tool_menu_item("Save Selected Node's Mesh As STL")
	if save_dialog:
		remove_child(save_dialog)
